from datetime import datetime

import datashader as ds
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import xarray as xr
from cftime import date2num, num2date
from matplotlib import cm
from matplotlib.colors import Normalize
from plotly.subplots import make_subplots


def spectrogram_preprocessing(
    xdf, log=False, dbscale=False, clip=[0.0, 1.0], canvas_dim=(400, 400)
):
    spec = xdf.data
    freq_dim = xdf.dims[0]
    freq = xdf[freq_dim].data
    dates = xdf["datetime"].data

    # set 0-frequency values to NaN
    if freq[0] == 0:
        freq = freq[1:]
        spec = spec[1:, :]

    vmin, vmax = clip
    _range = float(np.nanmax(spec) - np.nanmin(spec))
    vmin = np.nanmin(spec) + vmin * _range
    vmax = np.nanmin(spec) + vmax * _range

    with np.errstate(invalid="ignore"):
        spec = np.where(spec > vmin, spec, vmin)
        spec = np.where(spec < vmax, spec, vmax)

    if dbscale:
        spec = 10 * np.log10(spec)

    if canvas_dim is not None:
        dates = date2num(
            dates.astype("datetime64[us]").astype(datetime),
            units="hours since 1970-01-01 00:00:00.0",
            calendar="gregorian",
        )
        da = xr.DataArray(spec, coords=[("frequency", freq), ("datetime", dates)])
        cvs = ds.Canvas(plot_width=canvas_dim[0], plot_height=canvas_dim[1])
        agg = cvs.raster(source=da)
        freq, d, spec = agg.coords["frequency"].data, agg.coords["datetime"], agg.data
        dates = num2date(
            d, units="hours since 1970-01-01 00:00:00.0", calendar="gregorian"
        )
    return dates, freq, spec


def plot_ssam_mpl(xdf, axes=None, cmap=cm.viridis, figsize=(12, 6), **kwargs):
    """
    Plot spectrogram data using matplotlib.

    :param xdf: An xarray dataset as returned by :class:`zizou.ssam.SSAM`
    :type xdf: :class:`xarray.Dataset`
    :param log: Logarithmic frequency axis if True, linear frequency axis
                otherwise.
    :type log: bool
    :param axes: Plot into given axes.
    :type axes: :class:`matplotlib.axes.Axes`
    :param dbscale: If True 10 * log10 of color values is taken, if False the
                    sqrt is taken.
    :type dbscale: bool
    :param cmap: Specify a custom colormap instance.
    :type cmap: :class:`matplotlib.colors.Colormap`
    :param clip: adjust colormap to clip at lower and/or upper end. The given
                 percentages of the amplitude range (linear or logarithmic depending
                 on option `dbscale`) are clipped.
    :type clip: [float, float]
    :param figsize: Horizontal and vertical dimension 
                    of :class:`matplotlib.figure.Figure`
    :type figsize: [float, float]
    
    >>> import numpy as np
    >>> from zizou.ssam import SSAM, test_signal
    >>> from zizou.visualise import plot_ssam_mpl
    >>> ft = SSAM(interval=60, per_lap=.9, smooth=10, \
                  frequencies=np.arange(0., 25.1, .1), \
                  timestamp='start')
    >>> _ = ft.compute(test_signal())
    >>> fig = plot_ssam_mpl(ft.feature)

    """

    dates, freq, spec = spectrogram_preprocessing(xdf, **kwargs)
    dates = mdates.date2num(dates)

    # db scale and remove zero/offset for amplitude
    clip = kwargs.get("clip", [0.0, 1.0])
    vmin, vmax = clip
    norm = Normalize(vmin, vmax, clip=True)
    # ignore warnings due to NaNs
    if not axes:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
    else:
        ax = axes

    # calculate half bin width
    halfbin_time = (dates[1] - dates[0]) / 2.0
    halfbin_freq = (freq[1] - freq[0]) / 2.0

    if kwargs.get("log", False):
        # pcolor expects one bin more at the right end
        freq = np.concatenate((freq, [freq[-1] + 2 * halfbin_freq]))
        dates = np.concatenate((dates, [dates[-1] + 2 * halfbin_time]))
        # center bin
        dates -= halfbin_time
        freq -= halfbin_freq
        # Log scaling for frequency values (y-axis)
        ax.set_yscale("log")
        # Plot times
        ps = ax.pcolormesh(dates, freq, spec, cmap=cmap)
    else:
        # this method is much much faster!
        spec = np.flipud(spec)
        # center bin
        extent = (
            dates[0] - halfbin_time,
            dates[-1] + halfbin_time,
            freq[0] - halfbin_freq,
            freq[-1] + halfbin_freq,
        )
        ps = ax.imshow(spec, interpolation="nearest", extent=extent, cmap=cmap)

    # set correct way of axis, whitespace before and after with window
    # length
    ax.axis("tight")
    ax.grid(False)

    if axes:
        return ax

    fig.colorbar(ps)
    ax.set_ylabel("Frequency [Hz]")
    date_format = mdates.DateFormatter("%Y-%m-%d %H:%M:%S")
    ax.xaxis.set_major_formatter(date_format)
    # x_ticks = np.datetime64(dates[0]) + ax.get_xticks().astype('timedelta64[s]')
    # x_tick_labels = x_ticks.astype('datetime64[h]')
    # ax.set_xticklabels(x_tick_labels, rotation=45)
    fig.autofmt_xdate()

    return fig


def multi_feature_plot_mpl(feature1, feature2, figsize=(12, 5), **kwargs):
    """
    Plot two features in one figure. 2D features have to be plotted first.

    Parameters
    ----------
    feature1 : xarray.DataArray
        First feature to plot. Can be 1D or 2D.
    feature2 : xarray.DataArray
        Second feature to plot. Has to be 1D.
    figsize : tuple, optional
        Figure size, by default (12, 5)

    Returns
    -------
    matplotlib.figure.Figure
        Figure with two y-axis.

    """
    fig, ax1 = plt.subplots(figsize=figsize)
    if len(feature1.values.shape) == 2:
        ax1 = plot_ssam_mpl(feature1, axes=ax1)
    else:
        ax1.plot(mdates.date2num(feature1.datetime), feature1.data, color="deepskyblue")
        ax1.set_ylabel(feature1.name, color="deepskyblue")
    ax2 = ax1.twinx()
    ax2.plot(mdates.date2num(feature2.datetime), feature2.data, color="black")
    date_format = mdates.DateFormatter("%Y-%m-%d %H:%M:%S")
    ax1.xaxis.set_major_formatter(date_format)
    fig.autofmt_xdate()
    fig.colorbar(ax1.get_images()[0], ax=ax1)
    return fig


def plot_ssam_plotly(xdf, cmap="Ice_r", new_fig=True, **kwargs):
    """
    Plot spectrogram data using matplotlib.

    :param data: An xarray dataset as returned by :class:`zizou.ssam.SSAM`
    :type data: :class:`xarray.Dataset`
    :param log: Logarithmic frequency axis if True, linear frequency axis
                otherwise.
    :type log: bool
    :param dbscale: If True 10 * log10 of color values is taken, if False the
                    sqrt is taken.
    :type dbscale: bool
    :param cmap: Specify a custom colormap instance from
                 `plotly.colors.sequential`.
    :type cmap: list
    :param clip: adjust colormap to clip at lower and/or upper end. The given
                 percentages of the amplitude range (linear or logarithmic
                 depending on option `dbscale`) are clipped.
    :type clip: [float, float]
    :param new_fig: Whether to return a figure or just a graph object.
    :type new_flag: boolean
    :param max_tpoints: Maximum number of time intervals. If the datasets
                        exceeds that number it will be resampled
                        to max_tpoints.
    :type max_tpoints: int

    >>> import numpy as np
    >>> from zizou.ssam import SSAM, test_signal
    >>> from zizou.visualise import plot_ssam_plotly
    >>> ft = SSAM(interval=60, per_lap=.9, smooth=10, \
                  frequencies=np.arange(0., 25.1, .1), \
                  timestamp='start')
    >>> _ = ft.compute(test_signal())
    >>> fig = plot_ssam_plotly(ft.feature)
    """
    dates, freq, spec = spectrogram_preprocessing(xdf, **kwargs)
    if not new_fig:
        return go.Heatmap(z=spec, x=dates, y=freq, colorscale=cmap)

    fig = go.Figure(data=go.Heatmap(z=spec, x=dates, y=freq, colorscale=cmap))

    fig.update_layout(xaxis_nticks=20, yaxis_title="Frequency [Hz]")

    if kwargs.get("log", False):
        fig.update_yaxes(type="log")

    return fig


def get_trace(data, colour, kws_ssam):
    """
    Return a plotly graphics object from a feature.

    :param data: Feature data to plot.
    :type data: :class:`xarray.DataArray`
    :param colour: Colour for line plots.
    :type colour: str
    :returns: A 1-D line plot or a 2D heatmap.
    :rtype: :class:`plotly.graph_objs.Scatter` or
            :class:`plotly.graph_objs.Figure`
    """
    if data.name == "ssam":
        return plot_ssam_plotly(data, new_fig=False, dbscale=True, **kws_ssam)
    elif data.name == "filterbank":
        return plot_ssam_plotly(
            data, new_fig=False, dbscale=True, canvas_dim=None, **kws_ssam
        )
    elif data.name == "sonogram":
        return plot_ssam_plotly(
            data, new_fig=False, dbscale=False, canvas_dim=None, **kws_ssam
        )
    else:
        # data = data.fillna(data.mean())
        dates = data.datetime.data
        dates = pd.to_datetime(dates).to_pydatetime()
        featureData = data.to_dataframe()[data.name].interpolate("index").values
        return go.Scatter(
            x=dates,
            y=featureData,
            opacity=0.8,
            line_color=colour,
            name="<b>{}</b>".format(str(data.name)),
        )


def multi_feature_plot(
    feature1, feature2, equal_yrange=False, rangeslider=False, log=False, kws_ssam=None
):
    """
    Return a combined time-series plot of two features.

    :param feature1: First feature to plot with y-scale on
                     the left.
    :type feature1: str
    :param feature2: Second feature to plot with y-scale on
                     the right.
    :type feature2: str
    :param equal_yrange: If True align yscales.
    :type equal_yrange: bool
    :param rangeslider: If True show secondary panel for zooming.
    :type rangeslider: bool
    :param kws_ssam: Keywords to be passed to
                     :class:`zizou.visualise.plot_ssam_plotly`
    :type kws_ssam: dict
    :returns: plotly figure that can be displayed in a Jupyter
              notebook or in a dash app.
    :rtype: :class:`plotly.graph_objs.Figure`
    """
    if kws_ssam is None:
        kws_ssam = dict()

    if feature2.name in ["ssam", "sonogram", "filterbank"]:
        # Always plot 2D features first
        feature_tmp = feature1
        feature1 = feature2
        feature2 = feature_tmp

    fig = make_subplots(specs=[[{"secondary_y": True}]])

    trace = get_trace(feature1, "deepskyblue", kws_ssam)
    fig.add_trace(trace, secondary_y=False)

    fig.update_yaxes(
        tickfont=dict(color="deepskyblue"), title_font=dict(color="deepskyblue")
    )
    if feature2.name not in ["ssam", "sonogram", "filterbank"]:
        if feature1.name in ["ssam", "sonogram", "filterbank"]:
            trace = get_trace(feature2, "darkred", kws_ssam)
            fig.add_trace(trace, secondary_y=True)
            fig.update_yaxes(
                secondary_y=True,
                tickfont=dict(color="darkred"),
                title_font=dict(color="darkred"),
            )
            fig.update_yaxes(
                tickfont=dict(color="#4d4db3"),
                title_font=dict(color="#4d4db3"),
                secondary_y=False,
            )
        else:
            trace = get_trace(feature2, "dimgray", kws_ssam)
            fig.add_trace(trace, secondary_y=True)
            fig.update_yaxes(
                secondary_y=True,
                tickfont=dict(color="dimgray"),
                title_font=dict(color="dimgray"),
            )

    fig.update_layout(
        title={
            "text": "Features time series",
            "y": 0.9,
            "x": 0.5,
            "xanchor": "center",
            "yanchor": "top",
        },
        height=500,
        xaxis_rangeslider_visible=rangeslider,
    )

    # Set y-axes titles
    fig.update_yaxes(
        title_text="<b>{}</b>".format(str(feature1.name)), secondary_y=False
    )
    fig.update_yaxes(
        title_text="<b>{}</b>".format(str(feature2.name)), secondary_y=True
    )

    if equal_yrange:
        if feature1.name in ["sonogram", "filterbank", "ssam"]:
            # Smallest positive number
            y = fig["data"][0].y
            y_min = y[y > 0][0]
            y_max = fig["data"][0].y.max()
            y_range = np.array([y_min, y_max])
            if log:
                y_range = np.log10(y_range)
            fig.update_yaxes(range=y_range)
        else:
            y_min = min(float(feature1.min()), float(feature2.min()))
            y_max = max(float(feature1.max()), float(feature2.max()))
            if y_max > y_min:
                fig.update_yaxes(range=[y_min, y_max])

    if log:
        fig.update_yaxes(type="log")

    return fig


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import os

import numpy as np
import pandas as pd
import xarray as xr
from torch.utils.data import DataLoader, random_split

from zizou import (
    AnomalyDetectionBaseClass,
    SliceBatchSampler,
    ZizouDataset,
    min_max_scale_netcdf,
)


def test_dataset(setup_dataset):
    sg = setup_dataset
    zs = ZizouDataset(sg, ["rsam", "ssam"])
    shapes = zs.get_shapes()
    assert shapes[1] == 4
    assert zs.resolution == 1 / 6.0
    assert zs[:][0].shape == (4, 4)
    np.testing.assert_array_equal(zs[0:2][0], zs[[0, 1]][0])
    np.testing.assert_array_equal(zs[(1, 3)][0], zs[np.array([1, 3])][0])
    assert zs.get_dates()[0] == pd.Timestamp("2023-01-01 00:30:00")


def test_dataloader(setup_dataset):
    sg = setup_dataset
    zd = ZizouDataset(sg, ["rsam", "ssam"])
    train_size = 3
    test_size = 1
    train_dataset, test_dataset = random_split(zd, [train_size, test_size])
    assert len(train_dataset) == 3
    assert len(test_dataset) == 1
    train_loader = DataLoader(train_dataset, batch_size=2, shuffle=True)
    for batch, (X, y) in enumerate(train_loader):
        if batch == 0:
            assert X.shape == (2, 4)
        elif batch == 1:
            assert X.shape == (4,)


def test_SliceBatchSampler():
    slices = list(SliceBatchSampler(range(10), 4, False))
    assert slices == [slice(0, 4, None), slice(4, 8, None), slice(8, 10, None)]
    slices = list(SliceBatchSampler(range(5, 10), 4, False))
    assert slices == [slice(5, 9, None), slice(9, 10, None)]


def test_min_max_scaling(setup_dataset):
    sg = setup_dataset
    for feature in ["rsam"]:
        fin = sg.feature_path(feature)
        fout = os.path.join(sg.path, f"{feature}_scaling_params.nc")
        scaling_params = min_max_scale_netcdf(fin, feature, fout)
        assert scaling_params["min"] == 4.0
        assert scaling_params["max"] == 7.0
        assert scaling_params["median"] == 5.0

    for feature in ["ssam"]:
        fin = sg.feature_path(feature)
        fout = os.path.join(sg.path, f"{feature}_scaling_params.nc")
        scaling_params = min_max_scale_netcdf(fin, feature, fout)
        assert scaling_params["min"] == [1.0, 4.0, 9.0]
        assert scaling_params["max"] == [3.0, 7.0, 11.0]
        assert scaling_params["median"] == [2.0, 5.0, 10.0]


def test_scaling(setup_dataset):
    sg = setup_dataset
    zs = ZizouDataset(sg, ["rsam", "ssam"])
    xda_1D = zs.load_and_scale_feature("rsam")
    np.testing.assert_array_equal(xda_1D.values, np.array([0.0, 1 / 3.0, 1 / 3.0, 1.0]))
    xda_2D = zs.load_and_scale_feature("ssam")
    np.testing.assert_array_equal(
        xda_2D.values,
        np.array(
            [
                [0.0, 1 / 2.0, 1 / 2.0, 1.0],
                [0.0, 1 / 3.0, 1 / 3.0, 1.0],
                [1 / 2.0, 0.0, 1 / 2.0, 1.0],
            ]
        ),
    )



import logging
import logging.config
from configparser import ConfigParser
from datetime import datetime

import pytest

from zizou import get_data, zizou_logging
from zizou.autoencoder import AutoEncoder

logger = logging.getLogger(__name__)


@pytest.mark.slow
def test_fit(setup_ac):
    rootdir, store1, store2, config = setup_ac
    ac = AutoEncoder(store1, configfile=config, batch_size=64)
    ac.clear()
    ac.fit()
    assert float(ac.loss_) - 1 < 0.5


@pytest.mark.slow
def test_transform(setup_ac):
    savedir, store1, store2, config = setup_ac
    starttime = datetime(2023, 1, 8, 0, 0, 0)
    endtime = datetime(2023, 1, 9, 0, 0, 0)
    store1.starttime = starttime
    store1.endtime = endtime
    ac = AutoEncoder(store1, configfile=config, batch_size=64)
    ac.clear()
    ac.fit_transform(starttime, endtime)
    store1.starttime = starttime
    store1.endtime = endtime
    ae = store1("autoencoder_embedding")
    acl = store1("autoencoder_cluster")
    acloss = store1("autoencoder_loss")
    assert ae.shape == (6, 144)
    assert acl.shape == (5, 144)
    assert acloss.shape == (144,)


@pytest.mark.slow
def test_early_stopping(setup_ac):
    savedir, store1, store2, config = setup_ac
    starttime = datetime(2023, 1, 8, 0, 0, 0)
    endtime = datetime(2023, 1, 9, 0, 0, 0)
    store1.starttime = starttime
    store1.endtime = endtime
    ac = AutoEncoder(store1, configfile=config, batch_size=64)
    ac.clear()
    ac.fit_transform(starttime, endtime)
    store1.starttime = starttime
    store1.endtime = endtime
    ae = store1("autoencoder_embedding")
    acl = store1("autoencoder_cluster")
    acloss = store1("autoencoder_loss")
    assert ae.shape == (6, 144)
    assert acl.shape == (5, 144)
    assert acloss.shape == (144,)


@pytest.mark.slow
def test_spectrogram(setup_ac):
    rootdir, store1, store2, config = setup_ac
    starttime = datetime(2023, 1, 8, 0, 0, 0)
    endtime = datetime(2023, 1, 9, 0, 0, 0)
    ac = AutoEncoder(store1, configfile=config, features=["filterbank"], batch_size=64)
    ac.clear()
    ac.fit_transform(starttime, endtime)
    assert float(ac.loss_) - 1 < 0.5



import copy
import datetime
import json
import logging
import os

import h5netcdf
import numpy as np
import pandas as pd
import torch
import xarray as xr
from torch.utils.data import BatchSampler, Dataset

logger = logging.getLogger(__name__)


def min_max_scale_netcdf(
    input_file, variable_name, scaling_params_file, chunk_size=144
):
    if os.path.isfile(scaling_params_file):
        with open(scaling_params_file, "r") as f:
            scaling_params = json.load(f)
        return scaling_params

    with xr.open_dataset(input_file, group="original") as ds:
        data = ds[variable_name]
        if len(data.sizes) > 1:
            min_value = list(data.min(axis=1).values)
            max_value = list(data.max(axis=1).values)
            median_value = list(data.median(axis=1).values)
        else:
            min_value = float(data.min())
            max_value = float(data.max())
            median_value = float(data.median())
    scaling_params = {"min": min_value, "max": max_value, "median": median_value}
    with open(scaling_params_file, "w") as f:
        json.dump(scaling_params, f)

    return scaling_params


class SliceBatchSampler(BatchSampler):
    def __init__(self, sampler, batch_size, drop_last):
        super().__init__(sampler, batch_size, drop_last)
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        # Generate all indices from the sampler
        indices = list(self.sampler)

        # Yield slices instead of list of indices
        for i in range(indices[0], indices[-1] + 1, self.batch_size):
            if self.drop_last and i + self.batch_size > indices[-1] + 1:
                break
            yield slice(i, min(i + self.batch_size, indices[-1] + 1))


class ZizouDataset(Dataset):
    def __init__(self, store, features, target_transform=None, feature_transforms=None):
        self.target_transform = target_transform
        self.feature_transforms = feature_transforms
        self.store = copy.copy(store)
        self.features = features
        self.resolution = None
        self.starttime = store.starttime
        self.endtime = store.endtime
        self.init_dataset()
        self.npts = int(
            (self.endtime - self.starttime) / datetime.timedelta(hours=self.resolution)
        )
        self.nfeatures = self.get_shapes()[1]
        self.dates = None

    def get_dates(self):
        return pd.date_range(
            start=self.store.starttime,
            end=self.store.endtime,
            freq=datetime.timedelta(hours=self.resolution),
        )

    def init_dataset(self):
        self.scaling_params = {}
        for feature in self.features:
            scaling_params_file = os.path.join(
                self.store.path, f"scaling_params_{feature}.json"
            )
            fin = self.store.feature_path(feature)
            self.scaling_params[feature] = min_max_scale_netcdf(
                fin, feature, scaling_params_file
            )
            self.get_resolution(fin)

    def get_resolution(self, fin):
        with h5netcdf.File(fin, "r") as f:
            _resolution = f["original"].attrs["resolution"]
            if self.resolution is None:
                self.resolution = _resolution
            else:
                if self.resolution != _resolution:
                    msg = "Resolution is not consistent.\n"
                    msg += f"Expected {self.resolution}, got {_resolution}.\n"
                    msg += f"Feature: {fin}\n"
                    raise ValueError(msg)

    def get_shapes(self):
        nsamples = None
        nfeatures = 0
        for feature in self.features:
            shape_dict = self.store.shape(feature)
            for key, value in shape_dict.items():
                if key != "datetime":
                    nfeatures += value
                else:
                    if len(shape_dict) == 1:
                        nfeatures += 1
                    if nsamples is None:
                        nsamples = value
                    else:
                        if nsamples != value:
                            msg = "Number of samples is not consistent."
                            msg += f"Expected {nsamples}, got {value}"
                            msg += f"for feature: {feature}"
                            logger.warning(msg)
        return (nsamples, nfeatures)

    def load_and_scale_feature(self, feature):
        data = self.store(feature)
        if len(data.shape) > 1:
            data = data.fillna(
                np.array(self.scaling_params[feature]["median"])[:, np.newaxis]
            )
            data = (
                data - np.array(self.scaling_params[feature]["min"])[:, np.newaxis]
            ) / (
                np.array(self.scaling_params[feature]["max"])[:, np.newaxis]
                - np.array(self.scaling_params[feature]["min"])[:, np.newaxis]
            )
        else:
            data = data.fillna(self.scaling_params[feature]["median"])
            data = (data - self.scaling_params[feature]["min"]) / (
                self.scaling_params[feature]["max"]
                - self.scaling_params[feature]["min"]
            )
        return data

    def load_data(self, starttime, endtime):
        features = []
        var_cnt = 0
        for _f in self.features:
            logger.debug(f"Reading feature {_f} between {starttime} and {endtime}.")
            self.store.starttime = starttime
            self.store.endtime = endtime
            feat = self.load_and_scale_feature(_f)
            if len(feat.shape) > 1:
                for c in range(feat.shape[0]):
                    da = feat[c].reset_coords(drop=True)
                    feat_name = f"variable_{var_cnt}"
                    var_cnt += 1
                    features.append(da.rename(feat_name))
            else:
                if self.feature_transforms is not None:
                    features.append(
                        self.feature_transforms[_f](feat.rename(f"variable_{var_cnt}"))
                    )
                else:
                    features.append(feat.rename(f"variable_{var_cnt}"))
                var_cnt += 1
        feats = xr.merge(features).to_array()
        self.dates = feats.datetime.values
        return feats.transpose("datetime", "variable").values

    def __len__(self):
        return self.npts

    def __getitems__(self, idx):
        return self.__getitem__(idx)

    def __getitem__(self, idx):
        if isinstance(idx, int) or isinstance(idx, np.int64):
            start = [int(idx)]
            stop = [int(idx)]
        elif isinstance(idx, slice):
            start = idx.start
            stop = idx.stop
            if start is None:
                start = 0
            else:
                start = idx.start
            if stop is None:
                stop = self.npts - 1
            elif stop >= self.npts:
                stop = self.npts - 1
            else:
                stop = idx.stop - 1
            assert stop >= start, "Stop index must be greater than start index."
            start = [start]
            stop = [stop]

        elif isinstance(idx, list) or isinstance(idx, tuple):
            start = list(idx)
            stop = list(idx)

        elif isinstance(idx, np.ndarray):
            start = idx.tolist()
            stop = idx.tolist()

        # Handle PyTorch tensor of indices
        elif isinstance(idx, torch.Tensor):
            if (
                idx.dtype == torch.long or idx.dtype == torch.int
            ):  # Ensure it's an index tensor
                start = idx.tolist()
                stop = idx.tolist()
        else:
            raise TypeError(f"Invalid index type: {type(idx)}")

        all_data = []
        for _st, _sp in zip(start, stop):
            date_start = self.starttime + datetime.timedelta(
                hours=_st * self.resolution
            )
            date_end = self.starttime + datetime.timedelta(hours=_sp * self.resolution)
            _data = self.load_data(date_start, date_end)
            all_data.append(_data)
        sample = np.array(all_data).squeeze()
        target = np.array(all_data).squeeze()
        if self.target_transform is not None:
            sample = self.target_transform(sample)
            target = self.target_transform(target)
        return sample, target


class AnomalyDetectionBaseClass(object):
    def __init__(self, features, stacks=None, transforms=None):
        self.features = features
        self.stacks = stacks if stacks is not None else dict()
        self.transforms = transforms if transforms is not None else dict()

    def fit(self, starttime, endtime):
        """
        Train the model.
        """
        msg = "'training' is not implemented yet"
        raise NotImplementedError(msg)

    def infer(self, starttime, endtime):
        """
        Use the model to infer.
        """
        msg = "'inference' is not implemented yet"
        raise NotImplementedError(msg)

    def write_model_parameters(self):
        """
        Write principal components or
        model weights to file
        """
        msg = "'write_model_parameters' is not implemented yet"
        raise NotImplementedError(msg)

    def read_model_parameters(self):
        """
        Read principal components or
        model weights from file
        """
        msg = "'read_model_parameters' is not implemented yet"
        raise NotImplementedError(msg)

    def write_hyperparameters(self):
        """
        Store things like random seed,
        learning rate, number of principal
        components etc.
        """
        msg = "'write_hyperparameters' is not implemented yet"
        raise NotImplementedError(msg)

    def compute_anomaly_index(self):
        """
        A scalar value between 0 and 1 indicating
        the degree of anomaly (1=most anomaluous).
        """
        msg = "'compute_anomaly_index' is not implemented yet"
        raise NotImplementedError(msg)



import unittest
from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from obspy import Trace, UTCDateTime
from obspy.signal.util import enframe
from tonik.utils import generate_test_data

from zizou.util import (
    apply_freq_filter,
    apply_hanning,
    demean,
    round_time,
    stack_feature,
    stride_windows,
    trace_window_data,
    trace_window_times,
    window_array,
)


def test_stack_feature():
    """
    Test the stack_features function.
    """
    data = generate_test_data(
        dim=1, ndays=1, tstart=datetime(2021, 4, 3), add_nans=False
    )
    stack_length = "1h"
    xdf = stack_feature(data, stack_length)
    assert np.alltrue(xdf.datetime == data.datetime)

    a1 = data["rsam"].loc[dict(datetime="2021-04-03T00")].values
    a2 = xdf["rsam"].loc[dict(datetime="2021-04-03T00")].values
    assert a2[5] == (a1[:6]).sum() / len(a1[:6])
    stack_len_seconds = 3600

    num_windows = int(stack_len_seconds / pd.Timedelta(xdf.interval).seconds)
    rsam = data["rsam"]

    # Check correct values
    rolling_mean = [
        np.nanmean(rsam.data[(ind - num_windows + 1) : ind + 1])
        for ind in np.arange(num_windows, rsam.shape[0])
    ]
    np.testing.assert_array_almost_equal(
        np.array(rolling_mean), xdf.rsam.data[num_windows:], 6
    )


def test_stride_windows():
    data = np.random.randint(0, 10, size=10)
    test_array = np.vstack((data[0:5], data[3:8]))
    ds = stride_windows(data, 5, 2)
    np.testing.assert_array_equal(test_array.T, ds)
    ds = stride_windows(data[0:8], 5, 2)
    np.testing.assert_array_equal(test_array.T, ds)
    # Because the returned array is just a
    # different view on the original array,
    # you can't assign to it
    with pytest.raises(ValueError):
        ds[0, 0] = 5

    # But because ds is just a view, changing the
    # original array will change ds is, maybe,
    # unexpected ways:
    data[3] = 100
    assert ds[0, 1] == 100


def test_obspy_enframe():
    """
    Compare stride_windows to obspy's enframe which doesn't
    use 'as_strided'.
    """
    data = np.random.randint(0, 10, size=10)
    ds = window_array(data, 4, 2, taper=False, padval=None)
    ds1, nwin, nowin = enframe(data, np.ones(4), 2)
    np.testing.assert_array_equal(ds.T, ds1)


def test_demean():
    a = np.arange(9.0)
    ds = stride_windows(a, 3, 2)
    np.testing.assert_array_equal(demean(ds, axis=0)[0], -np.ones(7))
    a[1] = np.nan
    np.testing.assert_array_equal(demean(ds, axis=0)[1, 1], -0.5)


def test_taper():
    a = np.arange(9)
    ds = stride_windows(a, 3, 1)
    dsw, wvals = apply_hanning(ds, return_window=True)
    np.testing.assert_array_equal(dsw[:, 0], np.array([0, 1.0, 0]))
    np.testing.assert_array_equal(dsw[1], ds[1])
    np.testing.assert_array_equal(wvals, np.array([0, 1, 0])[:, np.newaxis])


def test_window_array():
    data = np.arange(10)
    # This will produce 4 windows with 5 elements each
    # overlapping by 3 elements; the last window will
    # contain one 0 to bring it to size 5
    res = window_array(data, 5, 3, remove_mean=False, taper=False, padval=0.0)
    assert res.shape == (5, 4)
    assert res[-1, -1] == 0
    # The returned array is just a different view of the
    # original so you can't assign to it
    with pytest.raises(ValueError):
        res[0, 0] = 100
    # Same as above but also remove the mean from each
    # column
    resm = window_array(data, 5, 3, remove_mean=True, taper=False, padval=0.0)
    np.testing.assert_array_equal(resm.sum(axis=0), np.zeros(resm.shape[1]))
    # The returned array is a new copy of the original
    # so you now can assign to it
    resm[0, 0] = 100
    assert resm[0, 0] == 100

    # Now remove the mean and also taper each window. This
    # is the default behaviour
    rest, wvals = window_array(data, 5, 3, padval=0.0)
    np.testing.assert_array_equal(rest[1], resm[1] * wvals[1])

    # This also works if we pad with np.nan
    rest, wvals = window_array(data, 5, 3, padval=np.nan)
    res_cp = res.copy()
    res_cp[-1, -1] = np.nan
    res_test = (res_cp[:4] - res_cp[:4, -1].sum() / 4) * wvals[:4]
    np.testing.assert_array_equal(rest[:4, -1], res_test[:4, -1])


def test_trace_window_data():
    tr = _dummy_trace_data(901)
    interval = 1.0
    result = list(trace_window_data(tr, interval))
    assert len(result) == 9
    np.testing.assert_array_equal(tr.data[100:201], result[1].data)
    with pytest.raises(ValueError):
        list(trace_window_times(tr, 100000))


def test_trace_window_times():
    n_samp = 901
    hz = 100.0
    tr = _dummy_trace_data(n_samp, hz)
    tr_len = (n_samp - 1) / hz

    # test simple case
    interval = 1.0
    n_windows = int(tr_len / interval)
    result = list(trace_window_times(tr, interval, interval))
    assert len(result) == n_windows
    assert tr.stats.starttime + 3 * interval == result[3][0]
    # test cases where final window len < interval
    interval = 1.05
    n_windows = int(tr_len / interval)
    result = list(trace_window_times(tr, interval, 1))
    assert len(result) == n_windows
    result = list(trace_window_times(tr, interval, 0.2))
    assert len(result) == n_windows + 1
    with pytest.raises(ValueError):
        list(trace_window_times(tr, 100000))


def test_apply_freq_filter():
    f = (2, 10)
    ftype = "bp"
    tr_untouched = _dummy_trace_data()
    tr1 = tr_untouched.copy()
    tr2 = tr_untouched.copy()
    apply_freq_filter(tr1, ftype, f)
    tr2.filter("bandpass", freqmin=f[0], freqmax=f[1], corners=4, zerophase=False)
    np.testing.assert_array_equal(tr1.data, tr2.data)


def test_round_time():
    """
    Test find closest time to interval.
    """
    time = UTCDateTime(2021, 5, 21, 13, 8)
    assert round_time(time, 600) == UTCDateTime(2021, 5, 21, 13, 10)


def _dummy_trace_data(n=2048, hz=100.0):
    t = np.linspace(0, 40 * np.pi, n) + np.linspace(0, 13 * np.pi, n)
    return Trace(
        np.sin(t), header={"starttime": "2020-01-01T12:00:00", "sampling_rate": hz}
    )


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np
import pytest
from obspy import UTCDateTime
from tonik import Storage

from zizou.spectral_features import SpectralFeatures
from zizou.util import test_signal


class SpectralTestCase(unittest.TestCase):
    def setUp(self):
        self.outdir = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.isdir(self.outdir):
            shutil.rmtree(self.outdir)

    sup = np.testing.suppress_warnings()
    sup.filter(module=np.matrixlib.defmatrix)  # module must match exactly

    def test_spectral(self):
        sf = SpectralFeatures(filtertype="bandpass", filterfreq=(0.5, 25))
        test_trace = test_signal(
            addchirp=False,
            frequencies=[5.0],
            amplitudes=[1.0],
            phases=[0.0],
            sinusoid=False,
            noise=False,
        )
        xds = sf.compute(test_trace)
        self.assertAlmostEqual(xds.predom_freq.data[0], 5, 0)
        self.assertAlmostEqual(xds.central_freq.data[0], 5, 0)
        self.assertAlmostEqual(xds.bandwidth.data[0], 0.062, 3)

    def test_spectral_with_gaps(self):
        sf = SpectralFeatures(filtertype="bandpass", filterfreq=(0.5, 25))
        test_trace = test_signal(
            addchirp=False,
            frequencies=[5.0],
            amplitudes=[1.0],
            phases=[0.0],
            sinusoid=False,
            noise=False,
            gaps=True,
        )
        xds = sf.compute(test_trace)
        self.assertAlmostEqual(xds.predom_freq.data[0], 5, 0)
        self.assertAlmostEqual(xds.central_freq.data[0], 5, 0)
        self.assertAlmostEqual(xds.bandwidth.data[0], 0.062, 3)
        self.assertEqual(np.isnan(xds.predom_freq.values).sum(), 3)

    @pytest.mark.slow
    def test_realtime(self):
        """
        Test with 1 hour windows as in the real-time computation
        """
        sf = SpectralFeatures(filtertype="highpass", filterfreq=(0.5, None))
        starttime = UTCDateTime(2020, 4, 21, 0, 0, 0)
        nsec = 86500  # Slightly longer than 24 h to avoid rounding issues
        step = 3600
        tr = test_signal(nsec=nsec, starttime=starttime, gaps=True)

        startwin = starttime
        sg = Storage("test", self.outdir)
        sg.starttime = starttime.datetime
        sg.endtime = (starttime + nsec).datetime
        st = sg.get_substore()
        while True:
            endwin = startwin + step
            if endwin > tr.stats.endtime:
                break
            ntr = tr.slice(startwin, endwin)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                st.save(sf.compute(ntr))
            startwin = endwin
        for feature in ["bandwidth", "central_freq", "predom_freq"]:
            self.assertEqual(st(feature).data.shape, (144,))



import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np
import pytest
from obspy import Trace, UTCDateTime
from scipy.signal import find_peaks
from tonik import Storage

from zizou.ssam import SSAM
from zizou.util import test_signal


@pytest.fixture(scope="module")
def setup_spectrogram():
    trace_ssam = test_signal(
        sinusoid=False, addchirp=False, frequencies=[1.0, 3.0, 10.0]
    )
    trace_ssam_gaps = test_signal(gaps=True)
    trace_sono = test_signal(
        offsets=[1000.0, 2000.0, 3000.0],
        frequencies=[1.0, 2.0, 5.0],
        amplitudes=[0.1, 1.0, 5.0],
        sinusoid=False,
        addchirp=False,
    )

    return trace_ssam, trace_sono, trace_ssam_gaps


def test_ssam(setup_spectrogram):
    trace, _, _ = setup_spectrogram
    test_peaks = np.array([1.0, 3.0, 10.0])
    s = SSAM(interval=60, per_lap=0.9, smooth=10)
    spec = s.compute(trace.copy())
    peaks = find_peaks(spec["ssam"].data[:, 0], height=0.1)[0]
    np.testing.assert_array_almost_equal(spec["frequency"].values[peaks], test_peaks, 2)


def test_sonogram(setup_spectrogram):
    _, trace, _ = setup_spectrogram
    s = SSAM(interval=60, per_lap=0.9, smooth=10)
    spec = s.compute(trace.copy())
    idx1 = int(
        np.abs(
            spec.datetime - np.datetime64((trace.stats.starttime + 1500.0).datetime)
        ).argmin()
    )
    idx2 = int(
        np.abs(
            spec.datetime - np.datetime64((trace.stats.starttime + 2600.0).datetime)
        ).argmin()
    )
    idx3 = int(
        np.abs(
            spec.datetime - np.datetime64((trace.stats.starttime + 3500.0).datetime)
        ).argmin()
    )
    assert (
        abs(
            float(
                spec.sonofrequency[spec["sonogram"].isel(dict(datetime=idx1)).argmax()]
            )
            - 2.295
        )
        < 1e-3
    )
    assert (
        abs(
            float(
                spec.sonofrequency[spec["sonogram"].isel(dict(datetime=idx2)).argmax()]
            )
            - 5.164
        )
        < 1e-3
    )
    assert (
        abs(
            float(
                spec.sonofrequency[spec["sonogram"].isel(dict(datetime=idx3)).argmax()]
            )
            - 11.62
        )
        < 1e-2
    )


def test_filterbank(setup_spectrogram):
    _, trace, _ = setup_spectrogram
    s = SSAM(interval=60, per_lap=0.9, smooth=10)
    spec = s.compute(trace.copy())
    idx1 = int(
        np.abs(
            spec.datetime - np.datetime64((trace.stats.starttime + 1500.0).datetime)
        ).argmin()
    )
    idx2 = int(
        np.abs(
            spec.datetime - np.datetime64((trace.stats.starttime + 2600.0).datetime)
        ).argmin()
    )
    idx3 = int(
        np.abs(
            spec.datetime - np.datetime64((trace.stats.starttime + 3500.0).datetime)
        ).argmin()
    )

    assert (
        abs(
            float(
                spec.fbfrequency[spec["filterbank"].isel(dict(datetime=idx1)).argmax()]
            )
            - 31.929
        )
        < 1e-3
    )
    assert (
        abs(
            float(
                spec.fbfrequency[spec["filterbank"].isel(dict(datetime=idx2)).argmax()]
            )
            - 2.165
        )
        < 1e-3
    )
    assert (
        abs(
            float(
                spec.fbfrequency[spec["filterbank"].isel(dict(datetime=idx3)).argmax()]
            )
            - 5.309
        )
        < 1e-3
    )


def test_ssam_with_gaps(setup_spectrogram):
    _, _, trace = setup_spectrogram
    linfreqs = np.arange(0, 25.1, 0.1)
    test_peaks = np.array([0.1, 3.0, 5, 6.1, 10.0])
    s = SSAM(interval=60, per_lap=0.9, smooth=10, frequencies=linfreqs)
    spec = s.compute(trace.copy())
    peaks = find_peaks(spec["ssam"].data[:, 0], height=1e-1)[0]
    np.testing.assert_array_almost_equal(linfreqs[peaks], test_peaks, 2)


@pytest.mark.slow
def test_realtime(tmp_path_factory):
    """
    Test with 1 hour windows as in the real-time computation
    """
    savedir = tmp_path_factory.mktemp("realtime")
    linfreqs = np.arange(0, 25.1, 0.1)
    testpeaks = np.array([0.1, 3.0, 5.0, 6.0, 10.0])
    s = SSAM(
        interval=60,
        frequencies=linfreqs,
        resample_int=(None, "10min"),
        timestamp="start",
    )
    starttime = UTCDateTime(2020, 4, 21, 0, 0, 0)
    nsec = 86500  # Slightly longer than 24 h to avoid rounding issues
    step = 3600
    tr = test_signal(nsec=nsec, starttime=starttime, gaps=True)
    sg = Storage(
        "test",
        savedir,
        starttime=starttime.datetime,
        endtime=(starttime + nsec).datetime,
    )
    store = sg.get_substore("WIZ", "00", "HHZ")
    startwin = starttime
    while True:
        endwin = startwin + step
        if endwin > tr.stats.endtime:
            break
        ntr = tr.slice(startwin, endwin)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            store.save(s.compute(ntr))
        startwin = endwin

    ssam = store("ssam")
    assert ssam.shape == (251, 144)
    peaks = find_peaks(ssam.data[:, 0], height=1e-1)[0]
    np.testing.assert_array_almost_equal(linfreqs[peaks], testpeaks, 2)


def test_config():
    """
    Test parsing of config parameters.
    """
    config = """
    default:
        interval: 600
    ssam:
        interval: 60
        per_lap: .9
        scale_by_freq: true
        timestamp: start
        frequencies:
        start: 0
        end: 25.1
        step: 0.1
        resample_int:
            upsample: null
            downsample: 10min
    """
    s = SSAM(configfile=config)
    assert s.resample_int[0] is None
    assert s.resample_int[1] == "10min"
    np.testing.assert_equal(s.frequencies, np.arange(0, 25.1, 0.1))



"""
Compute the displacement seismic amplitude ratio (DSAR).
"""

import logging

import numpy as np
import pandas as pd
import xarray as xr
import yaml
from obspy import UTCDateTime

from .feature_base import FeatureBaseClass
from .util import round_time, trace_window_times

logger = logging.getLogger(__name__)


class DSAR(FeatureBaseClass):
    """
    DSAR is the ratio of
    low frequency to high frequency amplitudes, as suggested in:
    https://doi.org/10.1130/G46107.1

    The time-domain displacement signal is bandpass filtered in two
    frequency bands, 4.5 to 8 Hz and 8 to 16 Hz, and DSAR is defined
    as the median ratio of the absolute signals. The quantity is
    interpreted in the publication as a change in seismic attenuation
    in the vicinity of the seismic station.
    """

    features = ["dsar"]

    def __init__(
        self,
        interval=600.0,
        filtertype="bandpass",
        lowerfreqband=(4.5, 8),
        higherfreqband=(8, 16),
        configfile=None,
        reindex=True,
    ):
        super(DSAR, self).__init__()
        self.interval = float(interval)
        self.filtertype = filtertype
        self.lowerfreqband = lowerfreqband
        self.higherfreqband = higherfreqband
        self.reindex = reindex

        if configfile is not None:
            try:
                with open(configfile, "r") as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            self.interval = c["default"].get("interval", interval)
            cd = c.get("dsar")
            if cd is not None:
                self.reindex = cd.get("reindex", reindex)
                self.filtertype = cd.get("filtertype", filtertype)
                lfreq = cd.get("lowerfreqband")
                if lfreq is not None:
                    self.lowerfreqband = (lfreq["low"], lfreq["high"])
                hfreq = cd.get("higherfreqband")
                if hfreq is not None:
                    self.higherfreqband = (hfreq["low"], hfreq["high"])

    def compute(self, trace):
        """
        Compute the displacement seismic amplitude ratio (DSAR).

        # TODO:
        Publication is quite vague and doesn't have enough detail
        to understand the method. Unclear whether median is every
        10 minutes or every day. This affects the bootstrap calc-
        ulation too, but this might not be necessary..

        :param stream: The SLCN code for the seismic stream
        :type stream: str
        :param startdate: Datetime of the first sample.
        :type startdate: :class:`obspy.UTCDateTime`
        :param enddate: Datetime of the last sample.
        :type enddate: :class:`obspy.UTCDateTime`
        :param interval: Length of the time interval in seconds over
                         which to compute DSAR (default 10 mins)
        :type interval: int, optional
        """
        if len(trace) < 1:
            raise ValueError("Trace is empty.")

        logger.info(
            "Computing DSAR for {} between {} and {}.".format(
                ".".join(
                    (
                        trace.stats.network,
                        trace.stats.station,
                        trace.stats.location,
                        trace.stats.channel,
                    )
                ),
                trace.stats.starttime.isoformat(),
                trace.stats.endtime.isoformat(),
            )
        )

        tr_lf = trace.copy()
        # handle gaps (i.e. NaNs)
        mdata = np.ma.masked_invalid(tr_lf.data)
        tr_lf.data = mdata
        # Get second trace for alternative filtering
        tr_hf = tr_lf.copy()

        # Filter settings
        lf_f1, lf_f2 = self.lowerfreqband
        hf_f1, hf_f2 = self.higherfreqband
        st_lf = tr_lf.split()
        st_hf = tr_hf.split()
        # Integrate to displacement
        st_lf.integrate("cumtrapz")
        st_hf.integrate("cumtrapz")

        st_lf.filter(
            self.filtertype, freqmin=lf_f1, freqmax=lf_f2, corners=4, zerophase=False
        )
        st_hf.filter(
            self.filtertype, freqmin=hf_f1, freqmax=hf_f2, corners=4, zerophase=False
        )

        starttime = trace.stats.starttime
        endtime = trace.stats.endtime
        st_lf.merge(fill_value=np.nan)
        st_lf.trim(starttime, endtime, pad=True, fill_value=np.nan)
        st_hf.merge(fill_value=np.nan)
        st_hf.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr_lf = st_lf[0]
        tr_hf = st_hf[0]

        # DSAR calculation
        feature_data = []
        feature_idx = []

        # loop through data in interval blocks
        for t0, t1 in trace_window_times(tr_lf, self.interval, min_len=0.8):
            tr_lf_int = tr_lf.slice(t0, t1)
            tr_hf_int = tr_hf.slice(t0, t1)

            absolute_lf = np.absolute(tr_lf_int.data)
            absolute_hf = np.absolute(tr_hf_int.data)

            # calculate ratio
            if np.isnan(absolute_lf).any() or np.isnan(absolute_hf).any():
                median = np.nan
            else:
                m = absolute_hf > 0
                ratio = absolute_lf[m] / absolute_hf[m]
                median = np.median(ratio)

            feature_data.append(median)
            feature_idx.append(t0.strftime("%Y-%m-%dT%H:%M:%S"))

        xda = xr.DataArray(
            feature_data, coords=[pd.DatetimeIndex(feature_idx)], dims="datetime"
        )
        xdf = xr.Dataset({"dsar": xda})
        xdf.attrs["starttime"] = trace.stats.starttime.isoformat()
        xdf.attrs["endtime"] = trace.stats.endtime.isoformat()
        xdf.attrs["station"] = trace.stats.station
        if self.reindex:
            starttime = UTCDateTime(str(xda.datetime.data[0]))
            starttime = round_time(starttime, self.interval)
            endtime = UTCDateTime(str(xda.datetime.data[-1]))
            endtime = round_time(endtime, self.interval)
            new_index = pd.date_range(
                starttime.datetime, endtime.datetime, freq="%dS" % int(self.interval)
            )
            xdf = xdf.reindex(dict(datetime=new_index), method="nearest")
        self.feature = xdf
        return self.feature
